""" BTC """
MINIMUM_BUY_SELL = .001
CURRENT_HIGHEST_USD = 42000

""" Coinbase """
# This is currently set to the actual fee rate, sometimes .0002 (.02%) is added as padding.
PRETEND_FEE_RATE = .005

# Actual fee rates
TAKER_FEE_RATE = .005
MAKER_FEE_RATE = .005

""" Margins """
# This is used with the fee rate. This allows a guaranteed profit if it reaches past fees and this extra rate
SELL_MIN = .002

""" Settings """
EMERGENCY_SELL_SAMPLE = 10



""" ZeroRPC server for communicating with other languages or GUIs """
from __future__ import print_function
from calc import calc as real_calc
import sys
import zerorpc
from Exchange import Exchange
from Coinbase_Pro.Coinbase_Pro import Coinbase_Pro
import Utils
import json


# A new process will be created for each total exchange. Bots can be appended to the predictor
class TradeInterface(object):
    def calc(self, text):
        """based on the input text, return the int result"""
        try:
            return real_calc(text)
        except Exception as e:
            return 0.0

    def echo(self, text):
        """echo any text """
        return text

    def init(self):
        self.__exchanges = []
        self.__utils = Utils.Utils()
        # Called from the dashboard
        try:
            with open('./Settings.json', 'r') as f:
                self.__user_preferences = json.load(f)
        # Must've been run in this folder
        except FileNotFoundError as e:
            with open('../../Tradingbot/Settings.json', 'r') as f:
                self.__user_preferences = json.load(f)

    def add_exchange(self, exchange_name, exchange_type, auth):
        if exchange_type == "coinbase_pro":
            self.__exchanges.append(Coinbase_Pro(exchange_name, self.__user_preferences, auth))
            return exchange_name
        else:
            Exception("Exchange type not found")

    """ External State """
    def get_exchange_state(self, name):
        for i in range(len(self.__exchanges)):
            if (self.__exchanges[i].get_name() == name):
                return [self.__exchanges[i].get_exchange_state(), name]

    """ 
    Internal State, this has all the currencies. This is mainly used for an initial definition of which currencies
    are being used, get model state is what will matter for the reporting into these blocks
    """
    def get_portfolio_state(self, name):
        for i in range(len(self.__exchanges)):
            if (self.__exchanges[i].get_name() == name):
                return [self.__exchanges[i].get_portfolio_state(), name]

    def run_model(self, name, currency=None):
        for i in range(len(self.__exchanges)):
            if (self.__exchanges[i].get_name() == name):
                return [self.__exchanges[i].start_models(currency), name]

    def assign_model(self, exchange_name, model_name, currency, args=None):
        for i in range(len(self.__exchanges)):
            if (self.__exchanges[i].get_name() == exchange_name):
                return [self.__exchanges[i].append_model(currency, args, model_name), exchange_name]

    """
    Three indicators at the top of the GUI - JSON with only 3 keys
    """
    def update_indicators(self, name):
        for i in range(len(self.__exchanges)):
            if (self.__exchanges[i].get_name() == name):
                return self.__exchanges[i].get_indicators()



def parse_port():
    port = 4242
    try:
        port = int(sys.argv[1])
    except Exception as e:
        pass
    return '{}'.format(port)

def main():
    addr = 'tcp://127.0.0.1:' + parse_port()
    s = zerorpc.Server(TradeInterface())
    s.bind(addr)
    print('start running on {}'.format(addr))
    s.run()

if __name__ == '__main__':
    main()


from Blankly.Coinbase_Pro.Coinbase_Pro_Tickers import Tickers as Coinbase_Pro_Ticker

import Blankly.Coinbase_Pro.Coinbase_Pro_Utils as Coinbase_Pro_Utils


class APIInterface:
    def __init__(self, type, authenticated_API):
        self.__type = type
        self.__calls = authenticated_API

    """
    Get all currencies in an account
    """
    def get_currencies(self, id=None):
        if self.__type == "coinbase_pro":
            if id == None:
                return self.__calls.getPortfolio()
            else:
                return self.__calls.getPortfolio("id")

    """
    Used for buying or selling market orders
    """
    def market_order(self, size, side, id):
        if self.__type == "coinbase_pro":
            order = Coinbase_Pro_Utils.CoinbaseProUtils().generate_market_order(size, side, id)
            # TODO exchange object needs to be generated here and then returned at some point, this invovles creating a ticker. Tickers are something that need to be managed carefully
            self.__calls.placeOrder(order)

    """
    Used for buying or selling limit orders
    """
    def limit_order(self, size, price, side, id):
        if self.__type == "coinbase_pro":
            order = Coinbase_Pro_Utils.CoinbaseProUtils().generate_limit_order(size, price, side, id)
            # TODO exchange object needs to be generated here and then returned at some point, this invovles creating a ticker. Tickers are something that need to be managed carefully
            self.__calls.placeOrder(order)

    """
    Creates ticker connection.
    """
    def create_ticker(self, callback, currency_id, log=""):
        if self.__type == "coinbase_pro":
            ticker = Coinbase_Pro_Ticker(currency_id, log=log)
            ticker.append_callback(callback)
            return ticker


import json
"""
This class creates a boilerplate exchange.
"""

class Exchange:
    def __init__(self, exchange_type, exchange_name, user_preferences):
        self.__name = exchange_name
        self.__type = exchange_type
        self.__preferences = user_preferences

    def get_name(self):
        return self.__name

    def getType(self):
        return self.__type

    def getPreferences(self):
        return self.__preferences


import json, numpy, time
from sklearn.linear_model import LinearRegression
import iso8601


class Utils:
    def __init__(self):
        pass

    """ 
    Json pretty printer for show arguments
    """
    def printJSON(self, jsonObject):
        print(self.returnPrettyJSON(jsonObject))

    """ 
    Json pretty printer for general string usage
    """
    def returnPrettyJSON(self, jsonObject):
        out = json.dumps(jsonObject.json(), indent=2)
        return out

    def getEpochFromISO8601(self, ISO8601):
        return time.mktime(iso8601.parse_date(ISO8601).timetuple())


    """
    Performs regression n points back
    """
    def getPriceDerivative(self, ticker, pointNumber):
        feed = numpy.array(ticker.getTickerFeed()).reshape(-1, 1)
        times = numpy.array(ticker.getTimeFeed()).reshape(-1, 1)
        if pointNumber > len(feed):
            pointNumber = len(feed)

        feed = feed[-pointNumber:]
        times = times[-pointNumber:]
        prices = []
        for i in range(pointNumber):
            prices.append(feed[i][0]["price"])
        prices = numpy.array(prices).reshape(-1, 1)

        regressor = LinearRegression()
        regressor.fit(times, prices)
        regressor.predict(times)
        return regressor.coef_[0][0]

    """ 
    Fit simple parabola 
    """
    def fitParabola(self, ticker, pointNumber):
        feed = ticker.getTickerFeed()
        times = ticker.getTimeFeed()
        if pointNumber > len(feed):
            pointNumber = len(feed)

        feed = feed[-pointNumber:]
        times = times[-pointNumber:]
        prices = []
        for i in range(pointNumber):
            prices.append(float(feed[i]["price"]))
            times[i] = float(times[i])

        # Pull the times back to x=0 so we can know what happens next
        latestTime = times[-1]
        for i in range(len(prices)):
            times[i] = times[i] - latestTime


        return numpy.polyfit(times, prices, 2, full=True)



import Keys
import Utils
import time
from Server import TradeInterface as TI
from Coinbase_Pro.Coinbase_Pro_API import API
from Coinbase_Pro.Coinbase_Pro import Coinbase_Pro


""" Define secret keys """
API_KEY = Keys.API_KEY
API_SECRET = Keys.API_SECRET
API_PASS = Keys.API_PASS


""" Define used currencies """
currencies = ["BTC", "XLM"]
market = "USD"

""" Define known best buy times """
buys = ["2:00"]
sells = ["21:00"]

""" Define libraries & Connections """
utils = Utils.Utils()
Server = TI()

# exchange = Coinbase_Pro("bill", {}, [API_KEY, API_SECRET, API_PASS])


# print("State:")
# exchange.append_model("COMP")
# print(exchange.get_portfolio_state())
# print("State with everything:")
# print("Started Model")
# exchange.start_models()
Server.init()
Server.add_exchange("API Portfolio", "coinbase_pro", [API_KEY, API_SECRET, API_PASS])
# calls = API(API_KEY, API_SECRET, API_PASS)
print(Server.get_portfolio_state("API Portfolio"))
Server.assign_model("API Portfolio", "Bill's Model", "COMP")
# print(calls.getPortfolio())
# print(Server.get_exchange_state("API Portfolio"))
Server.run_model("API Portfolio")
while True:
    print(Server.get_portfolio_state("API Portfolio"))
    time.sleep(1)

sys.exit()





bitcoinTicker = Tickers("BTC-USD", show=False)
time.sleep(2)

"""you always quit as soon as the going gets tough"""

print(call.getPortfolio())
print(call.getAccountInfo('GRT-USD'))
sys.exit()


ids = 0
def nextId():
    global ids
    ids += 1
    return {
        "id": ids
    }


manager = ProfitManager("BTC", bitcoinTicker)


# fit = utils.fitParabola(bitcoinTicker, 10000)
manager.addExchange(Exchange.Exchange(utils.generateMarketOrder(.001, "buy", "BTC-USD"), bitcoinTicker))

# Main thread becomes I/O thread
while True:
    command = input("View LocalAccount (v) or sell exchange 0 (s):")
    if command == "v":
        print("USD: " + str(LocalAccount.account["USD"]))
        print("BTC: " + str(LocalAccount.account["BTC-USD"]))
    elif command == "s":
        manager.getExchange(0).sellSelf()
        print("Selling all: ")
        print("USD: " + str(LocalAccount.account["USD"]))
        print("BTC: " + str(LocalAccount.account["BTC-USD"]))
    else:
        print("Command not found")

sys.exit()
print("Fee:")
exchange.getFee(show=True)
print("Value:")
print(exchange.getValueAtTime())
print("Profitable")
exchange.getProfitableSellPrice(show=True)
sys.exit()

# print(utils.getPriceDerivative(bitcoinTicker, 1000))
#
# while True:
#     time.sleep(1)
#
# order = utils.generateLimitOrder(Constants.MINIMUM_TO_SET_LIMIT, 15000, "buy", "BTC-USD")
# response, exchange = call.placeOrder(order, bitcoinTicker, show=True)
# print("Placed order")
# time.sleep(5)
# orders = call.getOpenOrders(show=True)
# print("Open Orders")
# print(orders)
# print("Canceling order")
# exchange.cancelOrder()
# time.sleep(3)
# print(exchange.confirmCanceled())
#
# while True:
#     time.sleep(1)


# order = utils.generateLimitOrder(Constants.MINIMUM_TO_SET_LIMIT, 15000, "buy", "BTC-USD")
# response, exchange = call.placeOrder(order, bitcoinTicker, show=True)

response = {
    'id': 3
}
call.getAccountInfo("BTC", "balance")
sys.exit()
exchange = Exchange.Exchange("buy", 46, bitcoinTicker, response, call)
while True:
    print(exchange.isProfitable(show=True))
    time.sleep(1)
    utils.getPriceDerivative(bitcoinTicker, 10)

""" Main trade loop """
scheduler = sched.scheduler(time.time, time.sleep)

def tradeFunction(id):
    print("started at " + str(time.time()))
    counter = 0
    while counter < 5:
        print("Trading " + id + ": " + str(counter))
        time.sleep(1)
        counter += 1
    print("finished at " + str(time.time()))


schedule.every().minute.at(":00").do(tradeFunction, "no1")
schedule.every().minute.at(":02").do(tradeFunction, "no1")

while True:
    schedule.run_pending()
    time.sleep(1)


"""
Class to manage a purchase lifecycle - a way to determine individual information about each buy and sell
    Copyright (C) <year>  <name of author>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import Constants, time, Local_Account.LocalAccount, Local_Account.Trade_Local as Trade_Local
from Utils import Utils
class Exchange:
    """ (Buying or selling (string), amount in currency (BTC/XLM), ticker object (so we can get time and value),
    limit if there is one) """
    def __init__(self, order, coinTicker, ApiCalls=None):
        self.__buyOrSell = order["side"]
        self.__purchaseTime = (coinTicker.getMostRecentTick())["time"]
        # Assigned below if there is an ApiCall attached
        self.__coinBaseId = None
        self.__calls = ApiCalls

        try:
            self.__limit = order["price"]
        except Exception as e:
            self.__limit = None


        self.__valueAtTime = float((coinTicker.getMostRecentTick())["price"])
        # This is the amount before fees
        self.__amountCurrency = order["size"]
        self.__ticker = coinTicker
        self.__active = True
        self.__utils = Utils()
        """ 
        self.__profitable is for any degree of profitability
        self.__reachedPastSellMin is if it's gone up a significant amount
        """
        self.__profitable = False
        self.__reachedPastMinToSell = False
        self.__sold = False
        if self.__calls is None:
            print("Buying locally")
            # Fees are calculated in the tradelocal function
            Trade_Local.tradeLocal(order["side"], self.__ticker.getCoinID(), order["size"], self.__ticker)
        else:
            response = self.__calls.placeOrder(order, self.__ticker, show=False)
            self.__coinBaseId = response["id"]

    def getPurchaseTime(self):
        return self.__purchaseTime

    def getSetLimit(self):
        return self.__limit

    """
    When these orders are used the amount of currency changes throughout the lifetime by the fee rate each time
    """
    def getAmountCurrencyExchanged(self):
        return self.__amountCurrency

    """
    On sell, we want to know how many dollars we got back.
    """
    def getDollarReturnOnSell(self):
        # TODO
        return None

    def getBoughtOrSold(self):
        return self.__buyOrSell

    def getValueAtTime(self):
        return self.__valueAtTime

    def getCoinBaseId(self):
        return self.__coinBaseId

    def getIfActive(self):
        return self.__active

    """ Returns true if the buy order can be sold at a profit or the sell order didn't loose money """
    def isProfitable(self, show=False):
        if self.__buyOrSell == "buy":
            if self.getProfitableSellPrice() < float(self.__ticker.getMostRecentTick()["price"]):
                if show:
                    print (True)
                return True
            else:
                if show:
                    print (False)
                return False
        else:
            if self.getProfitableSellPrice() > float(self.__ticker.getMostRecentTick()["price"]):
                if show:
                    print (True)
                return True
            else:
                if show:
                    print (False)
                return False

        # Old algorithm that didn't take into account the change in amount actually bought because of the fee
        # print("Before fees: " + str((self.__valueAtTime * self.__amountCurrency)))
        # then = (self.__valueAtTime * self.__amountCurrency)
        # now = (float((self.__ticker.getMostRecentTick())["price"]) * self.__amountCurrency)
        # if show:
        #     print("Before fees now: " + str(now))
        #     print("Before fees then: " + str(then))
        #     print("Difference with fees: " + str(((now - (now * Constants.PRETEND_FEE_RATE)) - (then - (then * Constants.PRETEND_FEE_RATE)))))
        #
        # if self.__buyOrSell == "buy":
        #     if ((now - (now * Constants.PRETEND_FEE_RATE)) - (then - (then * Constants.PRETEND_FEE_RATE))) > 0:
        #         return True
        #     else:
        #         return False
        # else:
        #     if ((then - (then * Constants.PRETEND_FEE_RATE)) - (now - (now * Constants.PRETEND_FEE_RATE))) > 0:
        #         return True
        #     else:
        #         return False

    def getFee(self, show=False):
        fee = (self.__valueAtTime * self.__amountCurrency) * Constants.PRETEND_FEE_RATE
        if (show):
            print (fee)
        return fee

    def cancelOrder(self):
        if self.__calls is not None:
            response = self.__calls.deleteOrder(self.__coinBaseId, show=False)
            try:
                if response["message"] == "Unauthorized.":
                    print("Not authorized to delete order: " + self.__coinBaseId)
            except TypeError as e:
                print("Canceled Order " + response)
            except Exception as e:
                print("FAILED to cancel order " + self.__coinBaseId)
            self.__active = False
        else:
            print("Cannot cancel order")

    def confirmCanceled(self):
        if self.__calls is not None:
            response = self.__calls.getOpenOrders()
            for i in range(len(response)):
                if (response[i]["id"] == self.__coinBaseId):
                    return False
            self.__active = False
            return True
        else:
            return "Cannot confirm cancellation"

    """ 
    Returns the price that a BUY needs to reach to be profitable
    Ths can be used to set limit orders
    """
    def getProfitableSellPrice(self, show=False):
        # price = self.__valueAtTime/(1 - Constants.PRETEND_FEE_RATE)
        # This one didn't work for the USD amount modified
        # price = self.__valueAtTime/(1 - (2 * Constants.PRETEND_FEE_RATE) + (Constants.PRETEND_FEE_RATE * Constants.PRETEND_FEE_RATE))
        price = ((Constants.PRETEND_FEE_RATE + 1)*(self.__valueAtTime)/(1-Constants.PRETEND_FEE_RATE))
        if show:
            print(price)
        return price

    """ 
    This allows this to be checked if this buy became profitable 
    """
    def inProfitableSellZone(self, show=False):
        self.__profitable = float((self.__ticker.getMostRecentTick()["price"])) > self.getProfitableSellPrice()
        if show:
            print(self.__profitable)
        return self.__profitable

    """ 
    Allows this exchange to be sold back in its entirety 
    """
    # TODO - This has turned into spaghetti code, fix it
    def sellSelf(self):
        if self.__calls is not None:
            # This doesn't need the fee calculation because that happens anyway
            self.__calls.placeOrder(self.__utils.generateMarketOrder(self.__amountCurrency, "sell", self.__ticker.getCoinID()))
            Trade_Local.tradeLocal("sell", self.__ticker.getCoinID(), self.__amountCurrency, self.__ticker)
        else:
            print("Selling locally only")
            # This one needs to include fees before and after
            self.__utils.tradeLocal("sell", self.__ticker.getCoinID(), self.__amountCurrency, self.__ticker)
        self.__sold = True


    def setIfPastSellMin(self, val):
        self.__reachedPastMinToSell = val

    def getIfPastSellMin(self, show=False):
        if show:
            print(self.__reachedPastMinToSell)
        return self.__reachedPastMinToSell

    def getIfSold(self):
        return self.__sold


"""
A calculator based on https://en.wikipedia.org/wiki/Reverse_Polish_notation
"""

from __future__ import print_function


def getPrec(c):
    if c in "+-":
        return 1
    if c in "*/":
        return 2
    if c in "^":
        return 3
    return 0


def getAssoc(c):
    if c in "+-*/":
        return "LEFT"
    if c in "^":
        return "RIGHT"
    return "LEFT"


def getBin(op, a, b):
    if op == '+':
        return a + b
    if op == '-':
        return a - b
    if op == '*':
        return a * b
    if op == '/':
        return a / b
    if op == '^':
        return a ** b
    return 0


def calc(s):
    numStk = []
    opStk = []
    i = 0
    isUnary = True
    while (i < len(s)):
        while (i < len(s) and s[i] == ' '):
            i += 1
        if (i >= len(s)):
            break
        if (s[i].isdigit()):
            num = ''
            while (i < len(s) and (s[i].isdigit() or s[i] == '.')):
                num += s[i]
                i += 1
            numStk.append(float(num))
            isUnary = False
            continue

        if (s[i] in "+-*/^"):
            if isUnary:
                opStk.append('#')
            else:
                while (len(opStk) > 0):
                    if ((getAssoc(s[i]) == "LEFT" and getPrec(s[i]) <= getPrec(opStk[-1])) or
                            (getAssoc(s[i]) == "RIGHT" and getPrec(s[i]) < getPrec(opStk[-1]))):
                        op = opStk.pop()
                        if op == '#':
                            numStk.append(-numStk.pop())
                        else:
                            b = numStk.pop()
                            a = numStk.pop()
                            numStk.append(getBin(op, a, b))
                        continue
                    break
                opStk.append(s[i])
            isUnary = True
        elif (s[i] == '('):
            opStk.append(s[i])
            isUnary = True
        else:
            while (len(opStk) > 0):
                op = opStk.pop()
                if (op == '('):
                    break
                if op == '#':
                    numStk.append(-numStk.pop())
                else:
                    b = numStk.pop()
                    a = numStk.pop()
                    numStk.append(getBin(op, a, b))
        i += 1

    while (len(opStk) > 0):
        op = opStk.pop()
        if op == '#':
            numStk.append(-numStk.pop())
        else:
            b = numStk.pop()
            a = numStk.pop()
            numStk.append(getBin(op, a, b))

    return numStk.pop()


# if __name__ == '__main__':
#     ss = [
#         "1 + 2 * 3 / 4 - 5 + - 6",  # -8.5
#         "10 + ( - 1 ) ^ 4",  # 11
#         "10 + - 1 ^ 4",  # 9
#         "10 + - - 1 ^ 4",  # 11
#         "10 + - ( - 1 ^ 4 )",  # 11
#         "5 * ( 10 - 9 )",  # 5
#         "1 + 2 * 3",  # 7
#         "4 ^ 3 ^ 2",  # 262144
#         "4 ^ - 3",  # 0.015625
#         "4 ^ ( - 3 )",  # 0.015625
#     ]
#     for s in ss:
#         res = calc(s)
#         print('{} = {}'.format(res, s))
